/*
 * stuart - ST2000DL003 userspace blockdevice over UART
 * Copyright (C) 2016, 2020 Petr Cvek
 *
 * based on BUSE example by:
 * Copyright (C) 2013 Adam Cozzette
 *
 * This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <poll.h>
#include <termios.h>

#include "buse.h"

/** *************************************************************************
WTF is this?
- userspace block device for linux (via NBD)
- compatible with ST2000DL003 drive with working ASCII (ctrl+z) console
- can READ all good sectors over UART
- these sectors are seen as a block device
- can be mounted (!)
- read HOWTO and BUG/TODO

HOWTO use:
- first you MUST be absolutely sure what this program does
- second you MUST be absolutely sure how seagate console works
- get a reliable working highspeed UART adapter
- research your buffer address and sector size (0x1040 = 8x(512 data + 2 crc + 6 garbage) )
	- used for "/FD405B8600,,,1040,1,"
- find your "User Data Base" start 4k sector (shown after reset)
- try every command used in this software with some terminal +(hexdump, xxd -g 1 -c 1)
	- /AR command uses "User Data Base" divided by 8 (4k -> 512 sectors)
- figure and redo the code for any different behavior (different firmware, congen configuration, drive failure)
- try change console baud speed (choose from /TB1), be sure your uart supports the speed
- wait only a few days to copy few gigabytes ;-)
- if you mount the device, do it READ ONLY and without any journaling
	- mount -t ext4 -o norecovery,ro,noload /dev/nbd0 /mnt/floppy
- use only one partition at time (START_512 and LEN_512)
	kernel will try to access every partition otherwise (takes a lot of time)
- !!! NO WARRANTIES !!!

BUG/TODO
- pthreads (1 buse main work, 1 buse reading, 1 serial ring filling)
	buse will duplicate uart file descriptor
- sometimes CRC fails, maybe bad drive or uart interferences?
- some uart blocking maybe redundant
- open uart in exclusive mode? (will do _funny_ stuff is concurent app is reading)
- in the current version you must have preset correct speed
- you need seagate console to be enabled
- written for x86-64 linux OS, different ARCHs not supported
- any error from drive will cause desynchronization and crash
- killing this program without unmounting can make kernel panic
- is it faster to /AR and /FD more sectors at once? (probably not)
- more parameters from argc/argv (debug)
- reverse engineer status bytes
- ESLIP speedup? (what are equivalent functions)
- usb dongles are SLOW!
- drive can't keep up will maximum host TX speed, you need to wait on echo

NOTICE useful informations:
speed benchmark
	iostat -k 10 /dev/nbd0
loop testing for interface BER
	while : ; do if [ -z "`pidof busexmp | cut -d ' ' -f 2`" ] ; then break; fi ;dd if=/dev/nbd0 bs=512 count=3000 iflag=direct | md5sum ; sleep 3 ; done
setting uart parameters (the code should do them by now automatically)
	stty -F /dev/ttyACM0 raw -echoe -echo -echoke speed 115200
print all available commands
	/CQ
using shell to sending the commands to the drive (at faster uart speeds, the will be overruns)
	echo -n -e "/AR13537f\n" > /dev/ttyACM0
	echo -n -e "/FD405B8600,,,1040,1,1\n" > /dev/ttyACM0
dump drive response
	xxd -g 1 -c 1 /dev/ttyACM0
ESLIP protocol thread
	https://forum.hddguru.com/viewtopic.php?f=1&t=38424
baudrate choices vs real speeds
	9600		TODO
	19200		TODO
	38400		38468.01144
	57600		TODO
	115200		TODO
	230400		TODO (may be interesting)
	460800		462093.86282
	625000		857150.51027
	921000		-//-
	921600		-//-
	1228000		1199985.00019
	1250000		-//-

CHANGELOG (newest):
- 2020.2
	reversed CRC algorithm
	basic recovery from interface error
	1200000 baud -> OS = 33kBps
	can be faster if you use hardware uart controller (USB will always wait 1ms between echoed characters)
- 2016 hddguru version
	/FD command + 460800 baud -> 34kBps
- original nonpublic version
	/2B command + 460800 bauds -> 4kBps
	possible speedup by using SymBits=20 (32bits per symbol)
* ************************************************************************* */

//user LBA section start (in 512B sectors, must be divided by 8 for /AR command)
#define BASE_512	0x00991590UL	//0x1322B2

/**
START_512 = first 512B sector to start
LEN_512 = number of 512B sectors
*/

/// WHOLE drive
// #define START_512	0UL
// #define LEN_512		(1927093140UL + 1979930925UL)

//! debug
// #define START_512	4
// #define LEN_512		100000

//!! test, only MBR
//    #define START_512	0
//    #define LEN_512	100000

//!! partition 1
//  #define START_512	63
//  #define LEN_512		41945652
//!! partition 2
//#define START_512	41945715
//#define LEN_512		41945715
//!! partition 3
 #define START_512	83891430
 #define LEN_512		1843201710
//!! partition 4
//#define START_512	1927093140
//#define LEN_512		1979930925

static char *portname = "/dev/ttyACM0";

#define BAUD_ORIG_HDD	38400
// #define BAUD_NEW_IFACE		460800
// #define BAUD_NEW_HDD		"460800"

#define BAUD_NEW_IFACE		1200000
#define BAUD_NEW_HDD		"1228000"

//read buffer for any response, must be enough for /FD length
//read it at once or you may got overruns
#define RAW_BUF_LEN	10000
unsigned char raw_rx_buf[RAW_BUF_LEN];
unsigned int raw_rx_buf_len = 0;
unsigned char sec_crc[2];

//NOTICE endians 0x1234 => "34 12"
//header before 512 sector dump
//ex: 07 00 07 00, 10 02 00 00, 17 00 01 00, 00 00 00 00
struct header {
	// /TO1 seems to be ESLIP oriented
	uint8_t unknown1[16];
	uint32_t addr;
	uint16_t len;
	uint8_t unknown2[2];
} __attribute__((packed));

/* **************** code starts ************ */
static int fd;	//uart global file descriptor
static int xmpl_debug = 1;	//debug for buse.c

int set_interface_attribs(int fd, unsigned int baud);
int get_interface_attribs(int fd);
int set_blocking(int fd, unsigned char min_bytes, unsigned char timeout);

static unsigned char ctrlz_ascii_cmd[] =
	"\x1a";
static unsigned char ctrlz_ascii_resp[] =
	"\x0d\x0a" "ASCII Diag mode" "\x0d\x0a\x0d\x0a" "F3 T>";

static unsigned char to_ascii_cmd[] =
	"/TO3";
static unsigned char to_ascii_resp[] =
	"\x0d\x0a" "Formatted ASCII Output Mode selected" "\x0d\x0a\x0d\x0a" "F3 T>";

static unsigned char to_bin_cmd[] =
	"/TO1";
static unsigned char to_bin_resp[] =
	"\x0d\x0a\x0d\x0a" "Raw Binary Output Mode selected" "\x0d\x0a\x0d\x0a" "F3 T>";

static unsigned char ctrlz_bin_cmd[] =
	"\x1a";
static unsigned char ctrlz_bin_resp[] =
	"\x0d\x0a" "ASCII Diag mode" "\x0d\x0a"
	"\x07\x00\x07\x00\x08\x00\x00\x00"
	"\x01\x00\x01\x00\x00\x00\x00\x00"
	"\x07\x00\x07\x00\x08\x00\x00\x00"
	"\x01\x00\x01\x00\x00\x00\x00\x00"
	"\x07\x00\x07\x00\x08\x00\x00\x00"
	"\x01\x00\x01\x00\x00\x00\x00\x00"
	"\x07\x00\x07\x00\x08\x00\x00\x00"
	"\x01\x00\x01\x00\x00\x00\x00\x00"
	"\x0d\x0a" "F3 T>";

//NOTICE /TB in /TO1 mode sometimes crashes
static unsigned char baud_ascii_cmd[] =
	"/TB" BAUD_NEW_HDD;
static unsigned char baud_ascii_resp[] =
	"\x0d\x0a" "F3 T>";

/*
there is a double packing:
	0x200 blocks with header
	512+8 sector (8 => 2B crc + garbage)
response for /TO1 (binary mode)
*/
static unsigned char dump_mem_cmd[] =
	"/FD405B8600,,,1040,1,";
static unsigned char dump_mem_resp[] =
	"\x0d\x0a";

#if 0	//NOTICE don't use
///2B2DC3,2DC3
static unsigned char dump_buf_cmd[] =
	"/2B2DC3,2DC3\n";
static unsigned char dump_buf_resp[] =
	"/2B2DC3,2DC3" "\x0d\x0a";
#endif

/**
/AR example:

"/AR1" 0d 0a

07 00 07 00 0c 00 00 00 26 00 01 00 00 00 00 00
00 00 00 00 07 00 07 00 b0 00 00 00 0b 00 02 00
00 00 00 00 01 00 00 00 80 00 00 00 cb 98 8b d8
81 06 f9 01 71 0a 00 00 00 00 00 00 00 00 ff ff

01 00 00 00 00 00 01 00 00 00 00 00 00 00 00 00
01 00 00 00 00 00 00 00 00 00 01 00 00 00 00 00
00 00 01 00 00 00 00 00 01 00 00 00 01 00 00 00
02 00 00 00 00 00 00 00 00 00 02 00 00 00 00 00

00 00 02 00 00 00 00 00 01 00 00 00 ff ff ff ff
00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
00 00 00 00 00 00 01 00 ff ff ff 00 00 00 00 00
00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
00 00 00 00 00 00 00 00 00 00 00 00

0d 0a "F3 A>"
*/

// static int ultradebug=1;

void dump_raw_buf(int len) {
#if 1
	fprintf(stderr,"dump raw buf (%d)\n", len);
	for (int idx=0;idx<len;idx++)
		fprintf(stderr,"%02hhx ", raw_rx_buf[idx] );
 	fprintf(stderr,"\n'");
 	for (int idx=0;idx<len;idx++)
 		fprintf(stderr,"%c", raw_rx_buf[idx]<0x20?'.':raw_rx_buf[idx]);
	fprintf(stderr,"'\n");
#endif
}

/// /////// seagate F3 crc16, use 16bit words //////

#define F3_CRC16_POLYNOM	0x2d

unsigned short seagate_crc16(
	unsigned short *d,
	unsigned int len,
	unsigned short pol,
	unsigned short init)
{
        unsigned int i;
        unsigned short sum = init;
        unsigned short pol_xor;

        for (i = 0; i < len; i++) {
                pol_xor = (sum & 0x8000) ? pol : 0;

                sum = (sum << 1) ^ d[i] ^ pol_xor;
        }
        return sum;
}


//flags = 1, wait on echo (flushes doesn't need these)
static int cmd_resp(
	int fd,
	unsigned char * cmd,
	unsigned int cmd_len,
	unsigned int expected_len,
	unsigned int flags
	)
{
	int ret;
	unsigned int sum;
        struct pollfd rd_poll;
        unsigned char readback;
        unsigned poll_timeout;
	unsigned char rb_test[100];

	if (!cmd) {
		fprintf(stderr, "!cr cmd NULL\n");
		return -901;
	}

	if (cmd_len == 0) {
		fprintf(stderr, "!cr zero cmd len\n");
		return -902;
	}

	if (expected_len > RAW_BUF_LEN) {
		fprintf(stderr, "!cr bad expected %i\n",expected_len);
		return -903;
	}

	sum = 0;

        rd_poll.fd = fd;
        rd_poll.events = POLLOUT;
        rd_poll.revents = 0;

        poll_timeout = 10;
	while (sum < cmd_len) {
		ret = poll(&rd_poll, 1, 200);
		if (ret == 0) {
			poll_timeout--;
			if (poll_timeout == 0) {
				fprintf(stderr, "!cr poll 1 timeouted\n");
				return -904;
			}
			continue;
		} else if (ret < 0) {
			fprintf(stderr, "!cr poll %i\n", errno);
			return errno;
		}

		ret = write(fd, &cmd[sum], 1);
		if (ret <= 0) {
			fprintf(stderr, "!cr write err %i\n", errno);
			return errno;
		}

		if (flags & 1) {
			set_blocking(fd, 1, 1);

			ret = read(fd, &readback, 1);
			if (ret <= 0) {
				fprintf(stderr, "!cr readback err %i\n", errno);
				return errno;
			}

			rb_test[sum] = readback;
			if (readback != cmd[sum]) {
				fprintf(stderr, "!cr cmp@%u %02hhx!=%02hhx '%s'\n", sum, readback, cmd[sum],cmd);
				rb_test[sum+1]=0;
				fprintf(stderr, "	'%s'\n", rb_test);
				//TODO abort, restart
			}
		}

		sum++;
	}

	if (flags & 1) {
		poll_timeout=10;
		while (1) {
			ret = poll(&rd_poll, 1, 200);
			if (ret == 0) {
				poll_timeout--;
				if (poll_timeout == 0) {
					fprintf(stderr, "!cr poll 2 timeouted\n");
					return -905;
				}
				continue;
			} else if (ret < 0) {
				fprintf(stderr, "!cr poll CR %i\n", errno);
				return errno;
			}
			break;
		}

		ret = write(fd, "\n", 1);
		if (ret <= 0) {
			fprintf(stderr, "!cr write CR err %i\n", errno);
			return errno;
		}
	}

	set_blocking(fd, 0, 2);

	//load buffer
	sum = 0;
	while(sum < expected_len) {
		//NOTICE using usb ACM CDC serial dongle, read returns
		//at the end of every packet
		ret = read(fd, &raw_rx_buf[sum], expected_len - sum);
		if (ret <= 0) {
			if (ret < 0) {
				fprintf(stderr, "!cr load, err %i %i\n", ret, errno);
				return ret;
			}
			break;
		} else {
			sum += ret;
		}
	}

// dump_raw_buf(sum);

	raw_rx_buf_len = sum;
	return sum;
}


/// ///////////// LRU cache implementation //////////////

//should be at least 9 (1x 4kB physical sector = 8 LBA)
//linux will use its own cache
#define CACHE_ENTRY_NUM		20

//newly added entry
#define CACHE_LRU_NEW		0

//oldest entry, will be erased soon
#define CACHE_LRU_OLDEST	100

//no entry, will be used first
#define CACHE_LRU_INVALID	-1

//counter for LRU policy, NOTICE signed
signed short cache_lru_counter[CACHE_ENTRY_NUM];

//address of 512B sectors
uint64_t cache_lba[CACHE_ENTRY_NUM];

//512B sector buffer
unsigned char cache_sector_data[CACHE_ENTRY_NUM][512];

static void cache_inc(void)
{
	for (int cache_idx = 0; cache_idx < CACHE_ENTRY_NUM; cache_idx++) {
		if ((cache_lru_counter[cache_idx] < CACHE_LRU_OLDEST) &&
		(cache_lru_counter[cache_idx] >= 0)){
			cache_lru_counter[cache_idx]++;
		}
	}
}

static int cache_find_lru(void)
{
	unsigned int found = 0;
	int max_val = -1;

	for (int cache_idx = 0; cache_idx < CACHE_ENTRY_NUM; cache_idx++) {
		if (cache_lru_counter[cache_idx] >= CACHE_LRU_OLDEST) {
			return cache_idx;
		} else if (cache_lru_counter[cache_idx] == CACHE_LRU_INVALID) {
			return cache_idx;
		}

		if (max_val < cache_lru_counter[cache_idx]) {
			max_val = cache_lru_counter[cache_idx];
			found = cache_idx;
		}
	}

	return found;
}

static int cache_find_lba(uint64_t lba)
{
	for (int cache_idx = 0; cache_idx < CACHE_ENTRY_NUM; cache_idx++) {
		if (cache_lru_counter[cache_idx] <= CACHE_LRU_OLDEST) {
			if (cache_lba[cache_idx] == lba) {
				return cache_idx;
			}
		}
	}

	//not found
	return -1;
}


static void cache_remove_entry(int cache_idx)
{
	if (cache_idx >= CACHE_ENTRY_NUM) {
		fprintf(stderr, "!cache rem index overflow %u\n",cache_idx);
		return;
	}

	cache_lba[cache_idx] = 0xdeadbeef;
	cache_lru_counter[cache_idx] = CACHE_LRU_INVALID;
}


static int cache_return_ptr(uint64_t lba, void** ptr)
{
	int cache_idx;

	//TODO overwrite/compare last value?
	if (cache_find_lba(lba) >= 0) {
		fprintf(stderr,"TODO already in cache\n");
		return -1;
	}

	cache_idx = cache_find_lru();

	cache_inc();

	cache_lba[cache_idx] = lba;

	cache_lru_counter[cache_idx] = CACHE_LRU_NEW;

	*ptr = (void*)&cache_sector_data[cache_idx][0];

	return cache_idx;
}


//NOTICE require valid entry
static void * cache_load(int cache_idx)
{
	//TODO invalid entry test
	if (cache_idx < 0) {
		fprintf(stderr, "!cache error %u\n",cache_idx);
		return NULL;

	} else if (cache_idx >= CACHE_ENTRY_NUM) {
		fprintf(stderr, "!cache rem index overflow %u\n",cache_idx);
		return NULL;
	}

 	cache_inc();
	cache_lru_counter[cache_idx] = CACHE_LRU_NEW;
	return &cache_sector_data[cache_idx][0];
}

static void cache_init(void)
{
	for (int cache_idx = 0; cache_idx < CACHE_ENTRY_NUM; cache_idx++) {
		cache_remove_entry(cache_idx);
	}
}

/// //////////////////////////////////////
/// //////////////// READ ////////////////
/// //////////////////////////////////////

static int xmp_read(
	void *buf,
	uint32_t len,
	uint64_t offset,
	void *userdata)
{
	unsigned int sector_byte_idx;
	unsigned int dump_idx;
	int ret;
	unsigned char *read_buf = buf;

//*userdata = int debug

	if ((offset % 512) != 0) {
		fprintf(stderr, "!!offset error %lx\n",offset);
		return -EIO;
	}

	if ((len % 512) != 0) {
		fprintf(stderr, "!!len error %x\n",len);
		return -EIO;
	}

	uint64_t lba = offset / 512;
	uint64_t lba_num = len / 512;
	uint64_t curr_lba;

// printf("r%lx/%lx: ", lba, lba_num);
// fflush(stdout);

	for (curr_lba = lba; curr_lba < (lba + lba_num); curr_lba++) {
		int cache_idx;

		cache_idx = cache_find_lba(curr_lba);

		if (cache_idx >= 0) {
			//found
			unsigned char * tbuf;
			tbuf = cache_load(cache_idx);
			if (tbuf == NULL) {
				fprintf(stderr,"!no buffer 1\n");
				return -EIO;
			}

			memcpy(read_buf, tbuf, 512);

			read_buf += 512;
		} else {
			//not found

			//max tries for 4k sector read
			#define MAX_TRIES	3
			unsigned int tries;

// printf("%lx ", curr_lba);

			/**
			NOTICE
			if the filesystem has "blocks" bigger than 512B,
			two HDD /AR accesses must be done

			+-------+---+-------------+
			|  phys |LBA|filesys block|
			|4k sect|512|1k,2k,4k,128k|
			+-------+---+-------------+
			|       | 0 | ---         |
			|       | 1 | ---         |
			|       | 2 | ---         |
			|   0   | 3 | 0 (4k block)|
			|       | 4 | 1           |
			|       | 5 | 2           |
			|       | 6 | 3           |
			|       | 7 | 4           |
			+-------+---+-------------+
			|       | 8 | 5           |
			|       | 9 | 6           |
			|   1   | a | 7           |
			|       | b | ---         |
			|       | c | ---         |
			+-------+---+-------------+
			*/

			uint64_t raw_4k_sector_addr;
			uint64_t first_raw_512_dumped;

			raw_4k_sector_addr =
				(BASE_512 + START_512 + curr_lba) / 8;

			//partition can start with some offset in raw 4k
			// NOTICE we will dump negative LBA
			first_raw_512_dumped = raw_4k_sector_addr * 8;

			unsigned char cmd_buf[20];
			int cmdlen;

			//run command
			cmdlen = snprintf(
				(char*)cmd_buf,
				sizeof(cmd_buf),
				"/AR%lx",
				raw_4k_sector_addr);

			if (cmdlen < 0) {
				fprintf(stderr,"!/AR%lx gen failed %i\n",
					raw_4k_sector_addr, cmdlen);
				return -EIO;
			}

			for (tries = 0; tries < MAX_TRIES; tries++) {
				ret = cmd_resp(fd, cmd_buf, cmdlen, 213, 1);
				if (ret <= 0) {
					fprintf(stderr,"!/AR%lx send failed %i\n",
						raw_4k_sector_addr,ret);
					dump_raw_buf(raw_rx_buf_len);
					cmd_resp(fd, "\x1b", 4, 5000, 0);
					continue;
				}

				ret = memcmp(&raw_rx_buf[0] , "\x0d\x0a", 2);
				if (ret) {
					fprintf(stderr,"!/AR%lx head cmp failed %i\n",
						raw_4k_sector_addr,ret);
					dump_raw_buf(raw_rx_buf_len);
					cmd_resp(fd, "\x1b", 4, 5000, 0);
					continue;
				}

				//TODO validate resp buffer?
				//offsets
				//0x50 ... sector 56 34 12 00 - LSB -> MSB
				//0x70 ... next sector??

				unsigned char prompt_a_resp[]="\x0d\x0a" "F3 A>";

#define RAWBUF_LEN (2+16*12+12)
				ret = memcmp(
					&raw_rx_buf[RAWBUF_LEN],
					prompt_a_resp,
					sizeof(prompt_a_resp)-1
				);
				if (ret) {
					fprintf(stderr,"!/AR%lx tail cmp failed %i (%i)\n",
					raw_4k_sector_addr,ret,RAWBUF_LEN);
					dump_raw_buf(raw_rx_buf_len);
					cmd_resp(fd, "\x1b", 4, 5000, 0);
					continue;
				}

				break;	//all OK
			}

			if (tries >= MAX_TRIES) {
				fprintf(stderr,"!/AR failed multiple times\n");
				cmd_resp(fd, "\x1b", 4, 5000, 0);
				return -EIO;
			}

			/// now dump 4k sector buffer

			/**
			!!TODO autodetect buffer address from "?" DBA column,
				"Default Diag Rd Buf" row
			dump 0x1040 bytes from previous read
			no "press any key to continue" paging
			packet size always 0x200 (padding)
			ESLIP packet header:

			from. to... length..... fcn.. rev.. ???????????
			07 00 07 00 10 02 00 00 17 00 01 00 00 00 00 00
			00  -   adress LSB
			86  -   adress
			5b  -   adress
			40  -   adress MSB
			11  -   segment length LSB
			01  -   segment length MSB
			01 00
			512 B sector data
			...
			next sector header
			...
			"\x0d\x0a" "F3 F>"
			*/

#define MEMDUMP_LEN	(4831 + sizeof(dump_mem_resp)-1)

			for (tries = 0; tries < MAX_TRIES; tries++) {
				/// sector dump
				ret = cmd_resp(fd, dump_mem_cmd, sizeof(dump_mem_cmd)-1, MEMDUMP_LEN, 1);
				if (ret <= 0) {
					fprintf(stderr,"!rd dump sect cmd failed %i\n",ret);
					dump_raw_buf(raw_rx_buf_len);
					cmd_resp(fd, "\x1b", 4, 5000, 0);
					continue;
				}

				if (ret != MEMDUMP_LEN) {
					fprintf(stderr,"!rd dump len W error %i, REDO for %i\n",ret,MEMDUMP_LEN);
					dump_raw_buf(raw_rx_buf_len);
					cmd_resp(fd, "\x1b", 4, 5000, 0);
					continue;
				}

				//first 2 bytes
				ret = memcmp(
					&raw_rx_buf[0],
					dump_mem_resp,
					sizeof(dump_mem_resp)-1
				);
				if (ret) {
					fprintf(stderr,"!rd dump resp failed %i\n",ret);
					dump_raw_buf(raw_rx_buf_len);
					cmd_resp(fd, "\x1b", 4, 5000, 0);
					continue;
				}

				unsigned char prompt_f_resp[]="\x0d\x0a" "F3 F>";

				ret = memcmp(
					&raw_rx_buf[MEMDUMP_LEN-(sizeof(prompt_f_resp)-1)],
					prompt_f_resp,
					sizeof(prompt_f_resp)-1
				);
				if (ret) {
					fprintf(stderr,"!rd prompt sync failed %i\n",ret);
					dump_raw_buf(raw_rx_buf_len);
					cmd_resp(fd, "\x1b", 4, 5000, 0);
					continue;
				}

				break;
			}

			sector_byte_idx = 0;

			//loop
			dump_idx = 0;

			unsigned int raw_idx;
			struct header *bin_hdr;

			//first dump header start
			raw_idx = sizeof(dump_mem_resp)-1;

			unsigned char *sect_ptr;
			sect_ptr = NULL;
			int curr_cache_idx=-1;

			do {
				//now dump sector buffer, NOTICE 4k physical sector

			// ultradebug=2;
				bin_hdr = (struct header *) &raw_rx_buf[raw_idx];
			// ultradebug=0;

				if ((bin_hdr->len != 0x200)&&(bin_hdr->len != 0x40)) {
					fprintf(stderr,"!dump unknown len %i\n", bin_hdr->len);
					cmd_resp(fd, "\x1b", 4, 5000, 0);
					//TODO segfault???
					return -EIO;
				}

				raw_idx += sizeof(struct header);

				for (int xxx = 0; xxx < bin_hdr->len; xxx++) {
					if (sector_byte_idx < 512) {
						if (sector_byte_idx == 0) {
							sect_ptr = NULL;

							//raw 4k partially extends outside defined user region
							//store only when not
							if ((first_raw_512_dumped + dump_idx) >=
								(BASE_512 + START_512 + 0)) {
								curr_cache_idx = cache_return_ptr(
										first_raw_512_dumped -
										BASE_512 -
										START_512 +
										dump_idx,
										(void*)&sect_ptr
										);
								}
						}

						if (sect_ptr) {
							sect_ptr[sector_byte_idx] = raw_rx_buf[raw_idx];
						}

						raw_idx++;
					} else if (sector_byte_idx < (512+2)) {
						//2 bytes 16bit crc16
						if (sect_ptr) {
							sec_crc[sector_byte_idx - 512] = raw_rx_buf[raw_idx];
						}
						raw_idx++;
					} else if (sector_byte_idx < (512+8)) {
						//512+2 to 512+8 is garbage
						raw_idx++;
					}

					sector_byte_idx++;
					if (sector_byte_idx >= (512+8)) {

						if (sect_ptr) {
							unsigned short mycrc;

							mycrc = seagate_crc16((void*)sect_ptr, 512/2, F3_CRC16_POLYNOM, 0);
							if (mycrc != ((sec_crc[1]<<8) | sec_crc[0])) {
								fprintf(stderr,"!crc doesn't match %02hhx-%02hhx != %04hx\n",sec_crc[0],sec_crc[1],mycrc);
								for (int iii=0;iii<512;iii++) {
									fprintf(stderr,"%02hhx ",sect_ptr[iii]);
								}
								fprintf(stderr,"\n");

fprintf(stderr," %i 0x%x-0x%x-0x%x+0x%x\n",curr_cache_idx, first_raw_512_dumped, BASE_512, START_512, dump_idx);

								cache_remove_entry(curr_cache_idx);
								cmd_resp(fd, "\x1b", 4, 4000, 0);
								//TODO segfault???
								return -EIO;
							}
						}

						sector_byte_idx = 0;
						dump_idx++;
					}
				}
			} while (dump_idx < 8);

			//rest of the last 0x200 block is unused

			cache_idx = cache_find_lba(curr_lba);
			if (cache_idx < 0) {
				fprintf(stderr,"!cache find failed after fill\n");
//dump cache
// for (int cache_idx = 0; cache_idx < CACHE_ENTRY_NUM; cache_idx++) {
// printf("%i: %x %i ... %x\n",cache_idx,cache_lba[cache_idx],cache_lru_counter[cache_idx],curr_lba);
// }
// 				cmd_resp(fd, "\x1b", 4, 4000, 0);

				return -EIO;
			}

			char * tbuf;
			tbuf = cache_load(cache_idx);
			if (tbuf == NULL) {
				fprintf(stderr,"!no buffer\n");
				cmd_resp(fd, "\x1b", 4, 4000, 0);

				return -EIO;
			}

			memcpy(read_buf, tbuf, 512);

			read_buf += 512;
		}
	}

// printf("\n");

	return 0;
}

//NOTICE no write implemented, ever :-P
static int xmp_write(
	const void *buf,
	uint32_t len,
	uint64_t offset,
	void *userdata)
{
	if (*(int *)userdata)
		fprintf(stderr, "WR o:%lx l:%x b:%p\n", offset, len, buf);
	return -EROFS;	//readonly EROFS
}

//TODO maybe send something like ^Z (0x1a)?
static void xmp_disc(void *userdata)
{
	(void)(userdata);
	fprintf(stderr, "Received a disconnect request.\n");
}

static int xmp_flush(void *userdata)
{
	(void)(userdata);
	fprintf(stderr, "Received a flush request.\n");
	return 0;
}

static int xmp_trim(uint64_t from, uint32_t len, void *userdata)
{
	(void)(userdata);
	fprintf(stderr, "T - %lu, %u\n", from, len);
	return 0;
}

static struct buse_operations aop = {
	.read = xmp_read,
	.write = xmp_write,
	.disc = xmp_disc,
	.flush = xmp_flush,
	.trim = xmp_trim,
	.block_size = 512,		//size of 1 block
	.block_count = LEN_512,		//number of blocks
};

/// //////////////////////////// MAIN /////////////////////////////////
/// //////////////////////////// MAIN /////////////////////////////////
/// //////////////////////////// MAIN /////////////////////////////////

int main(int argc, char *argv[])
{
	int ret;

	if (argc != 2) {
		printf(
			"STUart usage:\n"
			"	sudo modprobe nbd\n"
			"	sudo %s /dev/nbd0\n"
			"	sudo mount -t ext4 -o norecovery,ro,noload /dev/nbd0 /mnt/floppy\n"
			, argv[0]);
		return 1;
	}

	//high priority
	nice(-8);

	printf("open port\n");
//  	fd = open(portname, O_RDWR | O_NOCTTY | O_SYNC | O_NONBLOCK);
// 	fd = open(portname, O_RDWR | O_NOCTTY | O_SYNC);
//   	fd = open(portname, O_RDWR | O_NOCTTY );
//     	fd = open(portname, O_RDWR | O_NOCTTY | O_SYNC | O_NONBLOCK);
    	fd = open(portname, O_RDWR );	//this one works best, TODO maybe exclusive
	if (fd < 0) {
		fprintf(stderr,"error %d opening %s: %s\n", errno, portname, strerror (errno));
		return -1;
		}

	int current_speed;

	current_speed = get_interface_attribs(fd);

	printf("===start===\n");

	printf("get iface %i\n", current_speed);

#if 0
//TODO
	if ((current_speed != BAUD_NEW_IFACE) &&
		(current_speed != BAUD_ORIG_HDD)) {
printf("set iface to default speed\n");
		//if not MY max speed, UART was probably messed, expect rebooted drive
 		ret = set_interface_attribs (fd, BAUD_NEW_IFACE);  // set speed, 8n1 (no parity)
		if (ret) {
			fprintf(stderr,"error %d setting UART speed1: %s\n", errno, strerror (errno));
			close(fd);
			return errno;
		}
	}
#endif

	printf("escape\n");
 	for (int aaa=0;aaa<3; aaa++) {
		ret = cmd_resp(fd, "\x1b", 4, 5000, 0);
		if (ret < 0) goto error;
	}

	printf("any mode ctrl+z + flush\n");
	ret = cmd_resp(
		fd,
		ctrlz_ascii_cmd,
		sizeof(ctrlz_ascii_cmd)-1,
		5000,
		0
	);
	if (ret < 0) goto error;

	printf("any mode ctrl+z + flush2\n");

	ret = cmd_resp(
		fd,
		ctrlz_ascii_cmd,
		sizeof(ctrlz_ascii_cmd)-1,
		5000,
		0
	);
	if (ret <= 0) goto error;

#if 0
//TODO del?
	printf("ascii format /TO3\n");
	ret = cmd_resp(fd,
		to_ascii_cmd, sizeof(to_ascii_cmd)-1,
		to_ascii_resp, sizeof(to_ascii_resp)-1);
	if (ret) goto error;

	printf("ascii ctrl+z 0x1a\n");
	ret = cmd_resp(fd,
		ctrlz_ascii_cmd, sizeof(ctrlz_ascii_cmd)-1,
		ctrlz_ascii_resp, sizeof(ctrlz_ascii_resp)-1);
	if (ret) goto error;

	//NOTICE must in be ascii mode or crash
	printf("in ascii (!) set new baud speed, flush remainders\n");
	ret = cmd_resp(fd,
		baud_ascii_cmd, sizeof(baud_ascii_cmd)-1,
		NULL, 1000);
	if (ret) goto error;

	//TODO switch speed
	printf("switch adapter to the speed %i\n", BAUD_NEW_IFACE);
	ret = set_interface_attribs(fd, BAUD_NEW_IFACE);  //set MY speed, 8n1 (no parity) in UART adapter
	if (ret) goto error;

	printf("ascii ctrl+z 0x1a\n");
	ret = cmd_resp(fd,
		ctrlz_ascii_cmd, sizeof(ctrlz_ascii_cmd)-1,
		ctrlz_ascii_resp, sizeof(ctrlz_ascii_resp)-1);
	if (ret) goto error;

#endif

	printf("rawbin /TO1\n");
	ret = cmd_resp(fd, to_bin_cmd, sizeof(to_bin_cmd)-1, 44, 1);
	if (ret <= 0) goto error;

	ret = memcmp(&raw_rx_buf[0], to_bin_resp, sizeof(to_bin_resp)-1);
	if (ret) goto error;

	printf("rawbin ctrl+z 0x1a\n");
	ret = cmd_resp(fd, ctrlz_bin_cmd, sizeof(ctrlz_bin_cmd)-1, 90, 0);
	if (ret <= 0) goto error;

	ret = memcmp(&raw_rx_buf[0], ctrlz_bin_resp, sizeof(ctrlz_bin_resp)-1);
	if (ret) goto error;

	printf("Userspace UART block driver started\n");

// return 0;

	cache_init();

	ret = buse_main(argv[1], &aop, (void *)&xmpl_debug);

	printf("buse main exited %i\n",ret);

	close(fd);

	return ret;
error:
	fprintf(stderr,"!failed during init ret=%i\n", ret);
	close(fd);
	return ret;
}
